"""
pip install -U pyxk -i https://pypi.org/simple
"""
from pyxk.lazy_loader import LazyLoader
from pyxk.AES.cryptor import AESCryptor
from pyxk.requests.session import Session

os = LazyLoader("os", globals(), "os")
importlib = LazyLoader("importlib", globals(), "importlib")



__all__ = [
    item.removesuffix(".py")
    for item in os.listdir(os.path.dirname(__file__))
    if item not in ("__pycache__", "__init__.py")
]

def __getattr__(name):
    # print(f"{__name__!r}.__getattr__: {name!r}")
    if name not in __all__:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    return importlib.import_module(f".{name}", package=__name__)

def __dir__():
    return __all__



"""
import pyxk.utils
"""
from pyxk.lazy_loader import LazyLoader

re = LazyLoader("re", globals(), "re")
os = LazyLoader("os", globals(), "os")
base64 = LazyLoader("base64", globals(), "base64")
difflib = LazyLoader("difflib", globals(), "difflib")
multidict = LazyLoader("multidict", globals(), "multidict")
functools = LazyLoader("functools", globals(), "functools")
collections = LazyLoader("collections", globals(), "collections")



def open_decorator(built_in_open):
    """
    内置函数 open 装饰器
    作用: 写或追加模式下 创建不存在的目录
    """
    @functools.wraps(built_in_open)
    def wrapper(file, mode="r", **kwargs):
        # 判断 mode 是否属于写或追加模
        # collections.Counter 统计可迭代对象 每项出现的次
        # itertools.product 求多个可迭代对象的笛卡尔积
        create_mode = [collections.Counter(i+j) for i in ("w", "a") for j in ("b", "+", "b+", "")]
        # 创建目录
        if isinstance(mode, str) and collections.Counter(mode) in create_mode:
            folder = os.path.dirname(file)
            if not os.path.isdir(folder):
                os.makedirs(folder)
        return built_in_open(file, mode, **kwargs)
    return wrapper
make_open = open_decorator(open)


# base64 数据类型 正则表达式判断
B64_RE_PATTERN   = re.compile(r"^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?$")
B64_RE_PATTERN_B = re.compile(rb"^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?$")

def is_base64(data: str or bytes) -> bool:
    """
    判断base64数据类型 return: bool
    """
    if isinstance(data, bytes):
        return bool(B64_RE_PATTERN_B.match(data))
    if isinstance(data, str):
        return bool(B64_RE_PATTERN.match(data))
    # str 或 bytes 以外类型返回 False
    return False


def tobytes_from_base64(data: str or bytes, encoding="UTF-8"):
    """
    base64数据类型 转化为bytes
    如果不为base64数据类型 则返回原始数据
    """
    if (
        not isinstance(data, (str, bytes))
        or not is_base64(data)
    ):
        return False, data
    if isinstance(data, str):
        data = data.encode(encoding)
    return True, base64.b64decode(data)


# User-Agnet
UA_ANDROID  = "Mozilla/5.0 (Linux; Android 11; Pixel 5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.91 Mobile Safari/537.36"
UA_WINDOWNS = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36"
UA_MAC      = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36"
UA_IPHONE   = "Mozilla/5.0 (iPhone; CPU iPhone OS 13_2_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Mobile/15E148 Safari/604.1"
UA_IPAD     = "Mozilla/5.0 (iPad; CPU OS 13_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/87.0.4280.77 Mobile/15E148 Safari/604.1"
UA_SYMBIAN  = "Mozilla/5.0 (Symbian/3; Series60/5.2 NokiaN8-00/012.002; Profile/MIDP-2.1 Configuration/CLDC-1.1 ) AppleWebKit/533.4 (KHTML, like Gecko) NokiaBrowser/7.3.0 Mobile Safari/533.4 3gpp-gba"
UA_APAD     = "Mozilla/5.0 (Linux; Android 11; Phh-Treble vanilla Build/RQ3A.211001.001;) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/90.0.4430.91 Safari/537.36"

UA_ALL = {
    "android" : UA_ANDROID,
    "windows" : UA_WINDOWNS,
    "mac"     : UA_MAC,
    "iphone"  : UA_IPHONE,
    "ipad"    : UA_IPAD,
    "symbian" : UA_SYMBIAN,
    "apad"    : UA_APAD
}


def get_user_agent(user_agent, overwrite=False):
    """
    获取 UserAgent
    """
    if not isinstance(user_agent, str):
        raise ValueError(
            f"\033[31minvalid user_agent: '{user_agent}', "
            f"type: '{type(user_agent).__name__}'\033[0m")
    # 重写
    if overwrite is True:
        return user_agent

    user_agent = difflib.get_close_matches(user_agent.lower(), UA_ALL, 1)
    if not user_agent:
        return UA_ANDROID
    return UA_ALL[user_agent[0]]


def default_headers():
    """
    默认 Headers
    """
    headers = (("User-Agent", get_user_agent("android")),)
    return multidict.CIMultiDict(headers)



"""
实现 Requests 基类封装
"""
from multidict import CIMultiDict

from yarl import URL
from pyxk.utils import default_headers



class BaseSesssion:
    """
    pip install requests
    对 requests 封装基类
    """

    # 默认超时时间
    DEFAULT_TIMEOUT = 7

    def __init__(
        self,
        base_url: str or URL=None,
        headers: dict or CIMultiDict=None,
        timeout: int or float=None,
        verify: bool=True,
    ):
        self._base_url = base_url
        self._headers  = headers
        self._timeout  = timeout
        self._verify   = bool(verify)
        self.__initialization_attr()


    def __initialization_attr(self):
        self.__init_base_url()
        self.__init_headers()
        self.__init_timeout()


    def __init_base_url(self):
        base_url = getattr(self, "_base_url", None)
        if not base_url:
            setattr(self, "_base_url", None)
            return

        # base_url 格式错误
        if not isinstance(base_url, (str, URL)):
            raise TypeError(
                "\033[31m'base_url' type must be a 'str' or 'URL', "
                f"not '{type(base_url).__name__}'\033[0m")

        # 只有绝对路径 url 才能是base_url
        base_url = URL(str(base_url))
        if base_url.is_absolute():
            setattr(self, "_base_url", base_url)
        else:
            setattr(self, "_base_url", None)


    def __init_headers(self, overwrite=False):
        raw_headers = CIMultiDict() if overwrite is True else default_headers()
        headers = getattr(self, "_headers", None)
        if headers:
            headers = CIMultiDict(headers)
        else:
            headers = CIMultiDict()
        raw_headers.update(headers)
        setattr(self, "_headers", raw_headers)


    def __init_timeout(self):
        timeout = getattr(self, "_timeout", self.DEFAULT_TIMEOUT)
        if (
            not isinstance(timeout, (int, float))
            or timeout <= 0
        ):
            timeout = self.DEFAULT_TIMEOUT
        setattr(self, "_timeout", timeout)


    def build_url(
        self, url: str or URL, raw: bool=False
    ) -> str or URL:
        """
        构建完整 url, base_url=None 将抛出错误
        raw=True 返回字符串
        """
        if isinstance(url, str):
            url = URL(url)

        elif not isinstance(url, URL):
            raise TypeError(
                "\033[31m'url' type must be a 'str' or 'URL', "
                f"not '{type(url).__name__}'\033[0m"
            )

        if not url.is_absolute():
            if self.base_url is None:
                raise ValueError(f"\033[31murl '{url}' not absolute\033[0m")
            url = self._base_url.join(url)
        # 原生数据
        if raw:
            url = str(url)
        return url


    @property
    def base_url(self):
        if not hasattr(self, "_base_url"):
            setattr(self, "_base_url", None)
        base_url = getattr(self, "_base_url")
        if base_url:
            base_url = str(base_url)
        return base_url

    @base_url.setter
    def base_url(self, value):
        setattr(self, "_base_url", value)
        self.__init_base_url()

    @property
    def headers(self):
        if not hasattr(self, "_headers"):
            setattr(self, "_headers", CIMultiDict())
        return getattr(self, "_headers")

    @headers.setter
    def headers(self, value):
        setattr(self, "_headers", value)
        self.__init_headers(overwrite=True)

    @property
    def timeout(self):
        if not hasattr(self, "_timeout"):
            setattr(self, "_timeout", self.DEFAULT_TIMEOUT)
        return getattr(self, "_timeout")

    @timeout.setter
    def timeout(self, value):
        setattr(self, "_timeout", value)
        self.__init_timeout()



"""
延迟加载实现
"""
from types import ModuleType
from importlib import import_module



class LazyLoader(ModuleType):
    """
    模块 延迟加载
    :params: local_name: 模块引用名称
    :params: parent_module_globals: 加载全局变量变量
    :params: import_name: 导入模块名称
    """

    def __init__(
        self, local_name, parent_module_globals, import_name,
    ):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        super().__init__(import_name)

    def _loader(self):
        module = import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, name):
        # print(f"LazyLoader.__getattr__: {name!r}")
        module = self._loader()
        return getattr(module, name)

    def __dir__(self):
        module = self._loader()
        return dir(module)



"""
实现 Requests 封装
"""
from rich.console import Console

from pyxk.requests.base import BaseSesssion
from pyxk.lazy_loader import LazyLoader

os = LazyLoader("os", globals(), "os")
sys = LazyLoader("sys", globals(), "sys")
time = LazyLoader("time", globals(), "time")
copy = LazyLoader("copy", globals(), "copy")
multidict = LazyLoader("multidict", globals(), "multidict")
warnings = LazyLoader("warnings", globals(), "warnings")
functools = LazyLoader("functools", globals(), "functools")

requests = LazyLoader("requests", globals(), "requests")
rich_progress = LazyLoader("rich_progress", globals(), "rich.progress")

pyxk_utils = LazyLoader("pyxk_utils", globals(), "pyxk.utils")



def capture_request_exc(function):
    """
    request装饰器: 用于捕获请求中的部分异常
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        max_retry, retry_info, retry_timeout = 12, {}, 0.5
        while True:
            try:
                response = function(*args, **kwargs)
                # 抛出部分请求状态码
                status = response.status_code
                if 403 <= status <= 410:
                    raise requests.exceptions.InvalidURL(
                        f"invalid url:\033[4;31m'{response.url}'\033[0m, "
                        f"status_code: \033[31m'{status}'\033[0m")
                return response

            # 超时异常 - 自动重试 直到达到最大重试次数抛出
            except requests.exceptions.Timeout:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                url    = exc_value.request.url
                reason = str(exc_value.args[0].reason.args[1])
                # 记录异常
                retry_info.setdefault(reason, 0)
                retry_info[reason] += 1
                # 达到最大重试次数
                if retry_info[reason] > max_retry:
                    raise
                warnings.warn(f"\033[33m'{url}'\033[0m {reason}\n")
                time.sleep(retry_timeout)

            # 无网络异常请求 - 死循环请求
            except requests.exceptions.ConnectionError:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                url    = exc_value.request.url
                reason = exc_value.args[0]
                if hasattr(reason, "reason"):
                    reason = reason.reason
                reason = str(reason).lower()
                error  = "No address associated with hostname".lower()
                if error not in reason:
                    raise
                reason = reason.rsplit(": ", 1)[-1]
                warnings.warn(
                    f"\033[33m'{url}' \033[31m{reason}\033[0m\n")
                time.sleep(retry_timeout)
    return wrapper


console = Console()

class Session(BaseSesssion):
    """
    pip install requests
    对 requests 的封装, 修改了默认 UserAgent
    """

    def __init__(
        self, base_url=None, headers=None, timeout=None, keep_alive=False, **kwargs
    ):
        self._session    = None
        self._keep_alive = bool(keep_alive)
        super().__init__(base_url, headers, timeout, **kwargs)


    @property
    def session(self):
        """
        获取 requests.Sesssion()
        """
        session = getattr(self, "_session", None)
        if not getattr(self, "_keep_alive", False):
            session = requests.Session()
            setattr(self, "_session", session)

        if not isinstance(session, requests.Session):
            session = requests.Session()
            setattr(self, "_session", session)

        return session


    def request(self, method, url, *args, filename=None, resume=False, **kwargs):
        """
        session = requests.Session()
        session.request
        新增参数 filename:
        """
        with console.status(f"[yellow b]Send Request:[/] [bright_blue u]{url}[/]"):
            response = self._request(method, url, *args, **kwargs)

        # 不保存文件
        if filename is None:
            return response

        # 响应流大小
        chunk_size, filesize = 1024, 0
        filename = os.path.abspath(filename)
        filemode = "wb"
        length   = response.headers.get("Content-Length")

        # 断点续传
        if os.path.isfile(filename) and resume is True:
            # 获取 filename 字节
            filesize = os.path.getsize(filename)
            filemode = "ab"
            kwargs.setdefault("headers", {})
            headers = kwargs["headers"]
            # 请求头添加 range 请求部分内容
            headers["Range"] = f"bytes={filesize}-"
            response = self._request(method, url, *args, **kwargs)
            # 状态码 416 - 超出值域
            if response.status_code == 416:
                return response
            length = response.headers.get("Content-Length")

        length = int(length) + filesize if str(length).isdigit() else None

        progress_custom_columns  = (
            rich_progress.SpinnerColumn("line"),
            rich_progress.TextColumn("[progress.description]{task.description}"),
            rich_progress.BarColumn(),
            rich_progress.TaskProgressColumn(),
            rich_progress.TimeElapsedColumn(),
        )

        # 保存文件
        with pyxk_utils.make_open(filename, filemode) as file:
            with rich_progress.Progress(
                *progress_custom_columns,
                console=console,
                # transient=True, # 完成后清除进度条
            ) as progress:
                task = progress.add_task(
                    f"[magenta b]{os.path.basename(filename)}[/]",
                    total=length
                )
                progress.update(task, advance=filesize)

                for chunk in response.iter_content(chunk_size=chunk_size):
                    file.write(chunk)
                    progress.update(task, advance=chunk_size)

        console.print(
            "[green b]File Saved Successfully[/]"
            f"[yellow] ->[/] [bright_blue u]'{filename}'[/]\n")

        return response


    @capture_request_exc
    def _request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
        user_agent=None,
    ):
        url = self.build_url(url, raw=True)
        raw_headers = copy.deepcopy(self.headers)

        # 更新UserAgent
        if headers:
            raw_headers.update(multidict.CIMultiDict(headers))

        # 设置UserAgent
        if user_agent is not None and user_agent:
            if isinstance(user_agent, str):
                user_agent = [user_agent, False]
            user_agent = pyxk_utils.get_user_agent(*user_agent)
            raw_headers["user-agent"] = user_agent

        # 请求 timeout
        if (
            not isinstance(timeout, (int, float))
            or timeout <= 0
        ):
            timeout = self.timeout

        # verify
        verify = bool(self._verify) \
            if verify is None else bool(verify)

        return self.session.request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=raw_headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json)


    def get(self, url, allow_redirects=True, **kwargs):
        return self.request("GET", url, allow_redirects=allow_redirects, **kwargs)

    def post(self, url, data=None, json=None, allow_redirects=True, **kwargs):
        return self.request("POST", url, data=data, json=json, allow_redirects=allow_redirects, **kwargs)

    def head(self, url, allow_redirects=False, **kwargs):
        return self.request("HEAD", url, allow_redirects=allow_redirects, **kwargs)

    def options(self, url, allow_redirects=True, **kwargs):
        return self.request("OPTIONS", url, allow_redirects=allow_redirects, **kwargs)

    def put(self, url, data=None, **kwargs):
        return self.request("PUT", url, data=data, **kwargs)

    def delete(self, url, **kwargs):
        return self.request("DELETE", url, **kwargs)

    def patch(self, url, data=None, **kwargs):
        return self.request("PATCH", url, data=data, **kwargs)
